#!/usr/bin/env node
import { Cache } from "@samual/lib/Cache"
import { assert } from "@samual/lib/assert"
import { countHackmudCharacters } from "@samual/lib/countHackmudCharacters"
import { writeFilePersistent } from "@samual/lib/writeFilePersistent"
import { writeFile, readFile } from "fs/promises"
import { homedir } from "os"
import { extname, basename, resolve, dirname, relative } from "path"
import { supportedExtensions } from "../constants.js"
import { generateTypeDeclaration } from "../generateTypeDeclaration.js"
import { pull } from "../pull.js"
import { syncMacros } from "../syncMacros.js"
import "@samual/lib/readDirectoryWithStats"
import "@samual/lib/copyFilePersistent"
const version = "0.20.4",
	options = new Map(),
	commands = [],
	userColours = new Cache(user => {
		let hash = 0
		for (const char of user) hash += (hash >> 1) + hash + "xi1_8ratvsw9hlbgm02y5zpdcn7uekof463qj".indexOf(char) + 1
		return [colourJ, colourK, colourM, colourW, colourL, colourB][hash % 6](user)
	}),
	log = message => console.log(colourS(message))
for (const argument of process.argv.slice(2))
	if ("-" == argument[0]) {
		const [key, valueRaw] = argument.split("=")
		let value = valueRaw
		if (value)
			if ("true" == value) value = !0
			else if ("false" == value) value = !1
			else {
				const number = Number(value)
				isFinite(number) && (value = number)
			}
		else value = !0
		if ("-" == argument[1]) options.set(key.slice(2), value)
		else for (const option of key.slice(1)) options.set(option, value)
	} else commands.push(argument)
if ("v" == commands[0] || "version" == commands[0] || options.get("version") || options.get("v")) {
	console.log(version)
	process.exit()
}
const pushModule = import("../push.js"),
	processScriptModule = import("../processScript/index.js"),
	watchModule = import("../watch.js"),
	chokidarModule = import("chokidar"),
	{ default: chalk } = await import("chalk"),
	colourA = chalk.rgb(255, 255, 255),
	colourB = chalk.rgb(202, 202, 202),
	colourC = chalk.rgb(155, 155, 155),
	colourD = chalk.rgb(255, 0, 0),
	colourJ = chalk.rgb(255, 244, 4),
	colourK = chalk.rgb(243, 249, 152),
	colourL = chalk.rgb(30, 255, 0),
	colourM = chalk.rgb(179, 255, 155),
	colourN = chalk.rgb(0, 255, 255),
	colourS = chalk.rgb(122, 178, 244),
	colourV = chalk.rgb(255, 0, 236),
	colourW = chalk.rgb(255, 150, 224)
if (options.get("help") || options.get("h")) {
	logHelp()
	process.exit()
}
let autoExit = !0
switch (commands[0]) {
	case "push":
		{
			const hackmudPath = getHackmudPath(),
				sourcePath = commands[1]
			if (!sourcePath) {
				logError("Must provide the directory to push from\n")
				logHelp()
				break
			}
			const scripts = commands.slice(2)
			if (scripts.length) {
				const invalidScript = scripts.find(
					script => !/^(?:[a-z_][a-z\d_]{0,24}|\*)\.(?:[a-z_][a-z\d_]{0,24}|\*)$/.test(script)
				)
				if (invalidScript) {
					logError(`Invalid script name: ${JSON.stringify(invalidScript)}\n`)
					logHelp()
					break
				}
			} else scripts.push("*.*")
			const optionsHasNoMinify = options.has("no-minify")
			if ((optionsHasNoMinify || options.has("skip-minify")) && options.has("mangle-names")) {
				logError(
					`Options ${colourN("--mangle-names")} and ${colourN(optionsHasNoMinify ? "--no-minify" : "--skip-minify")} are incompatible\n`
				)
				logHelp()
				break
			}
			const shouldSkipMinify = options.get("no-minify") || options.get("skip-minify")
			let shouldMinify
			if (null != shouldSkipMinify) {
				if ("boolean" != typeof shouldSkipMinify) {
					logError(
						`The value for ${colourN(optionsHasNoMinify ? "--no-minify" : "--skip-minify")} must be ${colourV("true")} or ${colourV("false")}\n`
					)
					logHelp()
					break
				}
				shouldMinify = !shouldSkipMinify
			}
			const shouldMangleNames = options.get("mangle-names")
			if (null != shouldMangleNames && "boolean" != typeof shouldMangleNames) {
				logError(
					`The value for ${colourN("--mangle-names")} must be ${colourV("true")} or ${colourV("false")}\n`
				)
				logHelp()
				break
			}
			const shouldforceQuineCheats = options.get("force-quine-cheats")
			if (null != shouldforceQuineCheats && "boolean" != typeof shouldforceQuineCheats) {
				logError(
					`The value for ${colourN("--force-quine-cheats")} must be ${colourV("true")} or ${colourV("false")}\n`
				)
				logHelp()
				break
			}
			const { push } = await pushModule
			;(
				await push(sourcePath, hackmudPath, {
					scripts,
					onPush: info => logInfo(info, hackmudPath),
					minify: shouldMinify,
					mangleNames: shouldMangleNames,
					forceQuineCheats: shouldforceQuineCheats
				})
			).length || logError("Could not find any scripts to push")
		}
		break
	case "dev":
	case "watch":
		{
			const hackmudPath = getHackmudPath(),
				sourcePath = commands[1]
			if (!sourcePath) {
				logError("Must provide the directory to watch\n")
				logHelp()
				break
			}
			const scripts = commands.slice(2)
			if (scripts.length) {
				const invalidScript = scripts.find(
					script => !/^(?:[a-z_][a-z\d_]{0,24}|\*)\.(?:[a-z_][a-z\d_]{0,24}|\*)$/.test(script)
				)
				if (invalidScript) {
					logError(`Invalid script name: ${JSON.stringify(invalidScript)}\n`)
					logHelp()
					break
				}
			} else scripts.push("*.*")
			const optionsHasNoMinify = options.has("no-minify")
			if ((optionsHasNoMinify || options.has("skip-minify")) && options.has("mangle-names")) {
				logError(
					`Options ${colourN("--mangle-names")} and ${colourN(optionsHasNoMinify ? "--no-minify" : "--skip-minify")} are incompatible\n`
				)
				logHelp()
				break
			}
			const shouldSkipMinify = options.get("no-minify") || options.get("skip-minify")
			let shouldMinify
			if (null != shouldSkipMinify) {
				if ("boolean" != typeof shouldSkipMinify) {
					logError(
						`The value for ${colourN(optionsHasNoMinify ? "--no-minify" : "--skip-minify")} must be ${colourV("true")} or ${colourV("false")}\n`
					)
					logHelp()
					break
				}
				shouldMinify = !shouldSkipMinify
			}
			const shouldMangleNames = options.get("mangle-names")
			if (null != shouldMangleNames && "boolean" != typeof shouldMangleNames) {
				logError(
					`The value for ${colourN("--mangle-names")} must be ${colourV("true")} or ${colourV("false")}\n`
				)
				logHelp()
				break
			}
			const shouldforceQuineCheats = options.get("force-quine-cheats")
			if (null != shouldforceQuineCheats && "boolean" != typeof shouldforceQuineCheats) {
				logError(
					`The value for ${colourN("--force-quine-cheats")} must be ${colourV("true")} or ${colourV("false")}\n`
				)
				logHelp()
				break
			}
			const { watch } = await watchModule
			watch(sourcePath, hackmudPath, {
				scripts,
				onPush: info => logInfo(info, hackmudPath),
				typeDeclarationPath: (
					options.get("type-declaration-path") ||
					options.get("type-declaration") ||
					options.get("dts") ||
					options.get("gen-types")
				)?.toString(),
				minify: shouldMinify,
				mangleNames: shouldMangleNames,
				onReady: () => log("Watching"),
				forceQuineCheats: shouldforceQuineCheats
			})
			autoExit = !1
		}
		break
	case "pull":
		{
			const hackmudPath = getHackmudPath(),
				script = commands[1]
			if (!script) {
				logError("Must provide the script to pull\n")
				logHelp()
				break
			}
			const sourcePath = commands[2] || "."
			await pull(sourcePath, hackmudPath, script).catch(error => {
				console.error(error)
				logError(`Something went wrong, did you forget to ${colourC("#down")} the script?`)
			})
		}
		break
	case "sync-macros":
		{
			const hackmudPath = getHackmudPath(),
				{ macrosSynced, usersSynced } = await syncMacros(hackmudPath)
			log(`Synced ${macrosSynced} macros to ${usersSynced} users`)
		}
		break
	case "generate-type-declaration":
	case "gen-type-declaration":
	case "gen-dts":
	case "gen-types":
		{
			const target = commands[1]
			if (!target) {
				logError("Must provide target directory\n")
				logHelp()
				break
			}
			const sourcePath = resolve(target),
				outputPath = commands[2] || "./player.d.ts",
				typeDeclaration = await generateTypeDeclaration(sourcePath, getHackmudPath())
			let typeDeclarationPath = resolve(outputPath)
			await writeFile(typeDeclarationPath, typeDeclaration).catch(error => {
				assert(error instanceof Error, "src/bin/hsm.ts:327:35")
				if ("EISDIR" != error.code) throw error
				typeDeclarationPath = resolve(typeDeclarationPath, "player.d.ts")
				return writeFile(typeDeclarationPath, typeDeclaration)
			})
			log("Wrote type declaration to " + chalk.bold(typeDeclarationPath))
		}
		break
	case "help":
	case "h":
		logHelp()
		break
	case "golf":
	case "minify":
		{
			const target = commands[1]
			if (!target) {
				logError("Must provide target\n")
				logHelp()
				break
			}
			const fileExtension = extname(target)
			if (!supportedExtensions.includes(fileExtension)) {
				logError(
					`Unsupported file extension "${chalk.bold(fileExtension)}"\nSupported extensions are "${supportedExtensions.map(extension => chalk.bold(extension)).join('", "')}"`
				)
				break
			}
			const { processScript } = await processScriptModule,
				fileBaseName = basename(target, fileExtension),
				fileBaseNameEndsWithDotSrc = fileBaseName.endsWith(".src"),
				scriptName = fileBaseNameEndsWithDotSrc ? fileBaseName.slice(0, -4) : fileBaseName,
				scriptUser =
					"scripts" == basename(resolve(target, "..")) && "hackmud" == basename(resolve(target, "../../..")) ?
						basename(resolve(target, "../.."))
					:	void 0,
				optionsHasNoMinify = options.has("no-minify")
			if ((optionsHasNoMinify || options.has("skip-minify")) && options.has("mangle-names")) {
				logError(
					`Options ${colourN("--mangle-names")} and ${colourN(optionsHasNoMinify ? "--no-minify" : "--skip-minify")} are incompatible\n`
				)
				logHelp()
				break
			}
			const mangleNames_ = options.get("mangle-names")
			if (null != mangleNames_ && "boolean" != typeof mangleNames_) {
				logError(
					`The value for ${colourN("--mangle-names")} must be ${colourV("true")} or ${colourV("false")}\n`
				)
				logHelp()
				break
			}
			const mangleNames = mangleNames_,
				forceQuineCheats_ = options.get("force-quine-cheats")
			if (null != forceQuineCheats_ && "boolean" != typeof forceQuineCheats_) {
				logError(
					`the value for ${colourN("--force-quine-cheats")} must be ${colourV("true")} or ${colourV("false")}\n`
				)
				logHelp()
				break
			}
			const forceQuineCheats = forceQuineCheats_
			let outputPath =
				commands[2] ||
				resolve(
					dirname(target),
					fileBaseNameEndsWithDotSrc ? scriptName + ".js"
					: ".js" == fileExtension ? fileBaseName + ".min.js"
					: fileBaseName + ".js"
				)
			const golfFile = () =>
				readFile(target, { encoding: "utf8" }).then(async source => {
					const timeStart = performance.now(),
						{ script, warnings } = await processScript(source, {
							minify: !(options.get("no-minify") || options.get("skip-minify")),
							scriptUser,
							scriptName,
							filePath: target,
							mangleNames,
							forceQuineCheats
						}),
						timeTook = performance.now() - timeStart
					for (const { message, line } of warnings)
						log(`Warning "${chalk.bold(message)}" on line ${chalk.bold(line + "")}`)
					await writeFilePersistent(outputPath, script)
						.catch(error => {
							if (!commands[2] || "EISDIR" != error.code) throw error
							outputPath = resolve(outputPath, basename(target, fileExtension) + ".js")
							return writeFilePersistent(outputPath, script)
						})
						.then(() =>
							log(
								`Wrote ${chalk.bold(countHackmudCharacters(script))} chars to ${chalk.bold(relative(".", outputPath))} | took ${Math.round(100 * timeTook) / 100}ms`
							)
						)
				})
			if (options.get("watch")) {
				const { watch: watchFile } = await chokidarModule
				watchFile(target, { awaitWriteFinish: { stabilityThreshold: 100 } })
					.on("ready", () => log("Watching " + target))
					.on("change", golfFile)
				autoExit = !1
			} else await golfFile()
		}
		break
	default:
		commands[0] && logError(`Unknown command: ${colourL(commands[0])}\n`)
		logHelp()
}
autoExit && process.exit()
function logHelp() {
	const pushCommandDescription = "Push scripts from a directory to hackmud user's scripts directories",
		forceQuineCheatsOptionDescription = `Force quine cheats on. Use ${colourN("--force-quine-cheats")}=${colourV("false")} to force off`,
		hackmudPathOption = `${colourN("--hackmud-path")}=${colourB("<path>")}\n    Override hackmud path`
	console.log(colourN("Version") + colourS(": ") + colourV(version))
	switch (commands[0]) {
		case "dev":
		case "watch":
		case "push":
			console.log(
				colourS(
					`\n${colourJ("push" == commands[0] ? pushCommandDescription : "Watch a directory and push a script when modified")}\n\n${colourA("Usage:")}\n${colourC("hsm")} ${colourL(commands[0])} ${colourB('<directory> ["<script user>.<script name>"...]')}\n\n${colourA("Arguments:")}\n${colourB("<directory>")}\n    The source directory containing your scripts\n${colourB("<script user>")}\n    A user to push script(s) to. Can be set to wild card (${colourV("*")}) which will try\n    and discover users to push to\n${colourB("<script name>")}\n    Name of a script to push. Can be set to wild card (${colourV("*")}) to find all scripts\n\n${colourA("Options:")}\n${colourN("--no-minify")}\n    Skip minification to produce a "readable" script\n${colourN("--mangle-names")}\n    Reduce character count further but lose function names in error call stacks\n${colourN("--force-quine-cheats")}\n    ${forceQuineCheatsOptionDescription}\n${hackmudPathOption}\n${"push" == commands[0] ? "" : `${colourN("--type-declaration-path")}=${colourB("<path>")}\n    Path to generate a type declaration file for the scripts\n`}\n${colourA("Examples:")}\n${colourC("hsm")} ${colourL(commands[0])} ${colourV("src")}\n    Pushes all scripts found in ${colourV("src")} folder to all users\n${colourC("hsm")} ${colourL(commands[0])} ${colourV("src")} ${colourC("foo")}${colourV(".")}${colourL("bar")}\n    Pushes a script named ${colourL("bar")} found in ${colourV("src")} folder to user ${userColours.get("foo")}\n${colourC("hsm")} ${colourL(commands[0])} ${colourV("src")} ${colourC("foo")}${colourV(".")}${colourL("bar")} ${colourC("baz")}${colourV(".")}${colourL("qux")}\n    Multiple can be specified\n${colourC("hsm")} ${colourL(commands[0])} ${colourV("src")} ${colourC("foo")}${colourV(".")}${colourL("*")}\n    Pushes all scripts found in ${colourV("src")} folder to user ${userColours.get("foo")}\n${colourC("hsm")} ${colourL(commands[0])} ${colourV("src")} ${colourC("*")}${colourV(".")}${colourL("foo")}\n    Pushes all scripts named ${colourL("foo")} found in ${colourV("src")} folder to all user\n${colourC("hsm")} ${colourL(commands[0])} ${colourV("src")} ${colourC("*")}${colourV(".")}${colourL("*")}\n    Pushes all scripts found in ${colourV("src")} folder to all users`
				)
			)
			break
		case "pull":
			console.log(
				colourS(
					`\n${colourJ("Pull a script a from a hackmud user's script directory")}\n\n${colourA("Usage:")}\n${colourC("hsm")} ${colourL(commands[0])} ${colourB("<script user>")}${colourV(".")}${colourB("<script name>")}\n\n${colourA("Options:")}\n${hackmudPathOption}`
				)
			)
			break
		case "minify":
		case "golf":
			console.log(
				colourS(
					`\n${colourJ("Minify a script file on the spot")}\n\n${colourA("Usage:")}\n${colourC("hsm")} ${colourL(commands[0])} ${colourB("<target> [output path]")}\n\n${colourA("Options:")}\n${colourN("--no-minify")}\n    Skip minification to produce a "readable" script\n${colourN("--mangle-names")}\n    Reduce character count further but lose function names in error call stacks\n${colourN("--force-quine-cheats")}\n    ${forceQuineCheatsOptionDescription}\n${colourN("--watch")}\n    Watch for changes`
				)
			)
			break
		case "generate-type-declaration":
		case "gen-type-declaration":
		case "gen-dts":
		case "gen-types":
			console.log(
				colourS(
					`${colourJ("Generate a type declaration file for a directory of scripts")}\n\n${colourA("Usage:")}\n${colourC("hsm")} ${colourL(commands[0])} ${colourB("<directory> [output path]")}\n\n${colourA("Options:")}\n${hackmudPathOption}`
				)
			)
			break
		case "sync-macros":
			console.log(
				colourS(
					`\n${colourJ("Sync macros across all hackmud users")}\n\n${colourA("Options:")}\n${hackmudPathOption}`
				)
			)
			break
		default:
			console.log(
				colourS(
					`\n${colourJ("Hackmud Script Manager")}\n\n${colourA("Commands:")}\n${colourL("push")}\n    ${pushCommandDescription}\n${colourL("dev")}\n    Watch a directory and push a script when modified\n${colourL("golf")}\n    Minify a script file on the spot\n${colourL("gen-dts")}\n    Generate a type declaration file for a directory of scripts\n${colourL("sync-macros")}\n    Sync macros across all hackmud users\n${colourL("pull")}\n    Pull a script a from a hackmud user's script directory`
				)
			)
	}
}
function logInfo({ path, users, characterCount, error }, hackmudPath) {
	path = relative(".", path)
	error ?
		logError(`Error "${chalk.bold(error.message)}" in ${chalk.bold(path)}`)
	:	log(
			`Pushed ${chalk.bold(path)} to ${users.map(user => chalk.bold(userColours.get(user))).join(", ")} | ${chalk.bold(characterCount + "")} chars | ${chalk.bold(resolve(hackmudPath, users[0], "scripts", basename(path, extname(path))) + ".js")}`
		)
}
function logError(message) {
	console.error(colourD(message))
	process.exitCode = 1
}
function getHackmudPath() {
	const hackmudPathOption = options.get("hackmud-path")
	if (null != hackmudPathOption && "string" != typeof hackmudPathOption) {
		logError(`Option ${colourN("--hackmud-path")} must be a string, got ${colourV(hackmudPathOption)}\n`)
		logHelp()
		process.exit(1)
	}
	return (
		hackmudPathOption ||
		process.env.HSM_HACKMUD_PATH ||
		("win32" == process.platform ? resolve(process.env.APPDATA, "hackmud") : resolve(homedir(), ".config/hackmud"))
	)
}
